import io
import tempfile
import zipfile

import joblib
import pandas as pd


DF = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
DF2 = pd.DataFrame({'a': ['a', 'b'], 'b': ['c', 'd']})


def default_zip_file(df, df2):
    # type: (pd.DataFrame, pd.DataFrame) -> bytes
    """Return zip file with two DF saved using joblib."""
    with io.BytesIO() as memory_file:
        with zipfile.ZipFile(memory_file, mode='w') as zfile:
            tmp = tempfile.NamedTemporaryFile()
            with open(tmp.name, mode='wb'):
                joblib.dump(df, tmp.name)
            with open(tmp.name, mode='rb') as f:
                content = f.read()
                zfile.writestr('df', content)
            tmp.close()

            tmp2 = tempfile.NamedTemporaryFile()
            with open(tmp2.name, mode='wb'):
                joblib.dump(df2, tmp2.name)
            with open(tmp2.name, mode='rb') as f:
                content = f.read()
                zfile.writestr('df2', content)
            tmp2.close()
        memory_file.seek(0)
        return memory_file.getvalue()



from toucan_data_sdk.utils.helpers import (
    check_params_columns_duplicate,
    ParamsValueError
)


def compute_cumsum(
    df,
    id_cols,
    reference_cols,
    value_cols,
    new_value_cols=None,
    cols_to_keep=None
):
    """
    Compute cumsum for a group of columns.
    - `id_cols` are the columns id to create each group,
    - `reference_cols` are the columns to order the cumsum,
    - `value_cols` are the columns to cumsum,
    - `new_value_cols` are the new columns with the result cumsum
    - `cols_to_keep` are other column to keep in the dataframe. This option can
     be used if there is only one row by group [id_cols + reference_cols]

    For example :

    MONTH  DAY NAME  VALUE  X
     1      1    A      1  lo
     2      1    A      1  lo
     2     15    A      1  la
     1     15    B      1  la

    The function `compute_cumsum` with the arguments :
            id_cols=['NAME']
            reference_cols=['MONTH','DAY']
            cumsum_cols=['VALUE']
            cols_to_keep=['X']
    give as a result :


    NAME  MONTH  DAY  X  VALUE
     A     1      1  lo      1
     A     2      1  la      2
     A     2     15  lo      3
     B     1     15  la      1


    Args:
        df (pd.DataFrame):
        id_cols (list(str)):
        reference_cols (list(str)):
        value_cols (list(str)):
        new_value_cols (list(str)):
        cols_to_keep (list(str)):
    """
    if cols_to_keep is None:
        cols_to_keep = []

    if new_value_cols is None:
        new_value_cols = value_cols
    if len(value_cols) != len(new_value_cols):
        raise ParamsValueError('`value_cols` and `new_value_cols` needs '
                               'to have the same number of elements')

    check_params_columns_duplicate(id_cols + reference_cols + cols_to_keep + value_cols)

    levels = list(range(0, len(id_cols)))

    df = df.groupby(id_cols + reference_cols + cols_to_keep).sum()
    df[new_value_cols] = df.groupby(level=levels)[value_cols].cumsum()

    return df.reset_index()



import json
from contextlib import suppress

import pandas as pd
from tabulate import tabulate
TABULATE_CONF = {'headers': 'keys', 'showindex': False, 'tablefmt': 'pipe'}


def model(df: pd.DataFrame) -> list:
    """Generate a 'model' for a dataframe, atm list of cols with additional info"""
    mo = []
    for c, d in zip(df.columns, df.dtypes):
        o = {'col': c, 'dtype': d,
             'unique': df[c].is_unique,
             'nullable': df[c].isnull().any(),
             'first': df[c].iloc[0]}
        with suppress(TypeError):
            o['first'] = json.dumps(o['first'])
        mo.append(o)
    return mo


def tabulate_model(mo: list) -> str:
    return pd.DataFrame(mo).pipe(tabulate, **TABULATE_CONF)


def tabulate_data_source(ds: dict) -> str:
    """Tabulate an entry from etl_config."""
    return pd.DataFrame([ds]).melt(var_name='field').pipe(tabulate, **TABULATE_CONF)


def generate_md(data_sources: list, dfs: dict, header: list = None) -> list:
    """Generate one markdown block for each data_source in etl_config"""
    out = []

    if header:
        out += header

    for ds in data_sources:
        out.append('\n'.join([
            f'## Domain: {ds["domain"]}',
            '### ETL config',
            f'{tabulate_data_source(ds)}',
            '### Columns',
            f'{tabulate_model(model(dfs[ds["domain"]]))}'
        ]))

    return out



def replace(df, column, dst_column=None, **kwargs):
    """
    Replace values of a column (uses pandas.Series.replace)
    Args:
        df (pd.DataFrame): DataFrame to transform
        column (str): name of the column containing values to replace
        dst_column (str): optional, name of the column which will contain replaced
                          values (same as "column" by default)

        Other parameters are directly forwarded to pandas.Series.replace.
    """
    dst_column = dst_column or column
    df.loc[:, dst_column] = df[column].replace(**kwargs)
    return df



import pandas as pd
import itertools


def combine_columns_aggregation(
    df,
    id_cols,
    cols_for_combination,
    agg_func='sum'
):
    """
    Aggregates data to reproduce "All" category for requester
    - `id_cols` are the columns id to group,
    - `cols_for_combination` is a dict with the colums corresponding to
       the filters as key and their default value as value
    - `agg_func` (optional) is the function to use for aggregating the data.
       Can be callable, string, dictionary, or list of string/callables (by default 'sum')
    """
    requesters_cols = list(cols_for_combination.keys())
    requester_combination = [
        list(item) for i in range(0, len(requesters_cols) + 1)
        for item in itertools.combinations(requesters_cols, i)]
    dfs_result = []
    for comb in requester_combination:
        df_tmp = df.groupby(id_cols + comb).agg(agg_func).reset_index()
        for key in (set(cols_for_combination.keys()) - set(comb)):
            df_tmp[key] = cols_for_combination[key]
        dfs_result.append(df_tmp)

    return pd.concat(dfs_result)



def percentage(df, new_column, column, group_cols=None):
    """
    Add a column to the dataframe according to the groupby logic on group_cols
    :param df: Dataframe
    :param new_column: name of the new column
    :param column: name of the desired column you need percentage on
    :param group_cols: (str | list of str) or None
    :return: df + the percentage column
    """
    if group_cols is None:
        df[new_column] = 100. * df[column] / sum(df[column])
    else:
        df[new_column] = 100. * df[column] / df.groupby(group_cols)[column].transform(sum)
    return df



import pandas as pd


def waterfall(df, date, value, start, end, upperGroup,  # noqa:C901
              insideGroup=None, breakdown=None):
    """
    Return a line for each bars of a waterfall chart, totals, groups, subgroups.
    Compute the variation and variation rate for each line.

    Args:
        upperGroup (dict)
            - id: name of the column that contains upperGroups unique IDs
            - label: not required, text displayed under each upperGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of upperGroups
        insideGroup (dict)
            - id: name of the column that contains insideGroups unique IDs
            - label: not required, text displayed under each insideGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of insideGroups
        date (str): name of the column that id the period of each lines
        value (str): name of the column that contains the vaue for each lines
        start (dict):
            - label: text displayed under the first master column
            - date: value in the date col that id lines for the first period
        end (dict):
            - label: text displayed under the last master column
            - date: value in the date col that id lines for the second period
        breakdown (list) : (not implemented)
        # fillValues (bool): (case when false is not implemented)

    TODO:
        * idea to implement breakdown, call waterfall() on data filtered by filter.
        * crash gracefully if data or config is not coherent
        * il y aura des cols en trop: les gérer
    """

    if len(df) == 0:
        return df

    if breakdown is not None:
        raise NotImplementedError(
            'We will add breakdown support on your request, please contact the devs')

    groups = {
        'upperGroup': {
            'type': 'parent',
            'id':  'upperGroup',
            'order': {
                'by': ['upperGroup_order', 'groups'],
                'ascending': [True, True]
            },
            'obj': upperGroup
        }
    }
    if insideGroup is not None:
        groups['insideGroup'] = {
            'type': 'child',
            'id': 'insideGroup',
            'order': {
                'by': ['type', 'insideGroup_order', 'label'],
                'ascending': [False, True, True]
            },
            'obj': insideGroup
        }
    # prepare the dataframe with standard column names
    df = _compute_rename(df, date, value, groups)

    agg_conf = {'value': sum}
    agg_conf.update({f'{col}_label': 'first' for col in groups.keys()})
    agg_conf.update({f'{col}_order': 'first' for col in groups.keys()})
    df = df.groupby(list(groups.keys()) + ['date']).agg(agg_conf).reset_index()

    df_start, df_end = _compute_start_end(df, start, end)

    df = _compute_value_diff(df, start, end, groups)

    middle = _compute_upper_group(df)
    if insideGroup is not None:
        middle = pd.concat([middle, _compute_inside_group(df)])

    ret = _compute_order(df_start, df_end, middle, groups)

    return ret


def _compute_rename(df, date, value, groups):
    df = df.rename(columns={
        date: 'date',
        value: 'value'
    })
    for g_name, g in groups.items():
        df = df.rename(columns={g['obj']['id']: g_name})
        if 'label' not in g['obj']:
            df[f'{g_name}_label'] = df[g_name]
        else:
            df.rename(columns={g['obj']['label']: f'{g_name}_label'}, inplace=True)
        if 'groupsOrder' not in g['obj']:
            df[f'{g_name}_order'] = pd.np.nan
        else:
            df.rename(columns={g['obj']['groupsOrder']: f'{g_name}_order'}, inplace=True)
    return df


def _compute_start_end(df, start, end):
    """
    Compute two dataframes with value for start and end
    Args:
        totals(dataframe):

    Returns: Dataframe, Dataframe

    """
    result = {}
    time_dict = {'start': start, 'end': end}
    totals = df.groupby('date').agg({'value': sum}).reset_index()
    for time_name, time in time_dict.items():
        if not totals[totals['date'] == time['id']].empty:
            value = totals.loc[
                totals['date'] == time['id'], 'value'
            ].values[0]
        else:
            value = 0
        result[time_name] = pd.DataFrame([{
            'value': value,
            'label': time['label'],
            'groups': time['label']
        }])
    return result['start'], result['end']


def _compute_value_diff(df, start, end, groups):
    """
    Compute diff value between start and end
    Args:
        df(dataframe):

    Returns: Dataframe

    """
    start_values = df[df['date'] == start['id']].copy()
    end_values = df[df['date'] == end['id']].copy()

    merge_on = []
    for key, group in groups.items():
        merge_on = merge_on + [key, f'{key}_label', f'{key}_order']

    df = start_values.merge(end_values,
                            on=merge_on,
                            how='outer',
                            suffixes=('_start', '_end'), )

    # necessary before calculating variation
    df[['value_start', 'value_end']] = df[['value_start', 'value_end']].fillna(0)
    df['value'] = df['value_end'] - df['value_start']
    df.drop(['date_start', 'date_end', 'value_end'], axis=1, inplace=True)
    df.rename(columns={'upperGroup': 'groups'}, inplace=True)
    return df


def _compute_inside_group(df):
    """
    Compute inside Group
    Args:
        df(dataframe):

    Returns: Dataframe

    """
    inside_group = df.copy()
    inside_group['type'] = 'child'
    inside_group['variation'] = inside_group['value'] / inside_group[
        'value_start']
    inside_group.drop(['upperGroup_label', 'insideGroup', 'value_start'],
                      axis=1, inplace=True)
    inside_group.rename(columns={'insideGroup_label': 'label'},
                        inplace=True)
    return inside_group


def _compute_upper_group(df):
    """
    Compute upperGroup
    Args:
        df (Dataframe):

    Returns: Dataframe

    """
    upper_group = df.groupby(['groups']).agg({
        'value': sum,
        'value_start': sum,
        'upperGroup_label': 'first',
        'upperGroup_order': 'first'
    }).reset_index()
    upper_group['type'] = 'parent'
    upper_group['variation'] = upper_group['value'] / upper_group[
        'value_start']
    upper_group.drop(['value_start'], axis=1, inplace=True)
    upper_group.rename(columns={'upperGroup_label': 'label'}, inplace=True)
    return upper_group


def _compute_order(df_start, df_end, df_middle, groups):
    order = {'by': [], 'ascending': []}
    for key, group in groups.items():
        order['by'] = order['by'] + group['order']['by']
        order['ascending'] = order['ascending'] + group['order']['ascending']

    df_middle = df_middle.sort_values(**order)

    ret = pd.concat([df_start, df_middle, df_end])

    for key, elt in groups.items():
        cond = ret['type'] == elt['type']
        ret.loc[cond, 'order'] = ret.loc[cond, f'{key}_order']
        ret.drop([f'{key}_order'], axis=1, inplace=True)
    return ret



import operator as _operator


def _basic_math_operation(df, new_column, column_1, column_2, op):
    """
    Basic mathematical operation to apply operator on `column_1` and `column_2`
    Both can be either a number or the name of a column of `df`
    Will create a new column named `new_column`
    """
    if not isinstance(column_1, (str, int, float)):
        raise TypeError(f'column_1 must be a string, an integer or a float')
    if not isinstance(column_2, (str, int, float)):
        raise TypeError(f'column_2 must be a string, an integer or a float')

    if isinstance(column_1, str):
        column_1 = df[column_1]
    if isinstance(column_2, str):
        column_2 = df[column_2]
    operator = getattr(_operator, op)
    df[new_column] = operator(column_1, column_2)
    return df


def add(df, new_column, column_1, column_2):
    """Add df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='add')


def subtract(df, new_column, column_1, column_2):
    """Subtract df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='sub')


def multiply(df, new_column, column_1, column_2):
    """Multiply df[value] (value: 'str') or value (number) and column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='mul')


def divide(df, new_column, column_1, column_2):
    """Divide df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='truediv')



import pandas as pd


def melt(df, id, value, dropna=False):
    """
    This function is useful to massage a DataFrame into a format where one or more columns
    are identifier variables (id), while all other columns,
    considered measured variables (value), are “unpivoted” to the row axis,
    leaving just two non-identifier columns, ‘variable’ and ‘value’.
    Args:
        df (pd.DataFrame): DataFrame to transform
        id (list): Column(s) to use as identifier variables
        value (list): Column(s) to unpivot.
        dropna (bool): dropna in added 'value' column
    """
    df = df[(id + value)]
    df = pd.melt(df, id_vars=id, value_vars=value)
    if dropna:
        df = df.dropna(subset=['value'])

    return df



# Spreadsheets doc:
# https://drive.google.com/drive/folders/0B56th7Lb-9vScy0tMlpIeGNxQ0E

from .add_missing_row import add_missing_row
from .combine_columns_aggregation import combine_columns_aggregation
from .clean import clean_dataframe
from .compute_cumsum import compute_cumsum
from .compute_evolution import (
    compute_evolution,
    compute_evolution_by_criteria,
    compute_evolution_by_frequency
)
from .compute_ffill_by_group import compute_ffill_by_group
from .roll_up import roll_up
from .two_values_melt import two_values_melt



def fillna(df, column, value):
    """
    Can fill NaN values from a column
    Args:
        df
        column
        value
    """
    if column in df.columns:
        df[column] = df[column].fillna(value)
    else:
        df[column] = value
    return df
